#include <gnome.h>
#include <libgnomevfs/gnome-vfs.h>
#include <glade/glade.h>

#include <utillib.h>

#include "project.h"
#include "project-properties.h"
#include "gdownman.h"
#include "new-url.h"
#include "new-project.h"


GType           gdownman_get_type           ();
static void     gdownman_class_init         (GdownmanClass *klass);
static void     gdownman_init               (Gdownman *gdownman);

static gint     delete_event_cb     (GtkWidget *window, GdkEventAny *e,
                                     gpointer data);

void    gdownman_quit               (GtkWidget *caller, gpointer data);
void    gdownman_select_all         (GtkWidget *caller, gpointer data);
void    gdownman_select_done        (GtkWidget *caller, gpointer data);
void    gdownman_select_stop        (GtkWidget *caller, gpointer data);
void    gdownman_select_queue       (GtkWidget *caller, gpointer data);
void    gdownman_select_active      (GtkWidget *caller, gpointer data);
void    gdownman_select_matched     (GtkWidget *caller, gpointer data);
void    gdownman_project_change     (GtkWidget *caller, GtkNotebookPage *page,
                                     guint page_num,
                                     gpointer data);
void    gdownman_project_add        (Gdownman *gdownman, Project *project);
void    gdownman_project_remove     (Gdownman *gdownman, Project *project);
void    gdownman_project_new        (GtkWidget *caller, gpointer data);
void    gdownman_project_delete     (GtkWidget *caller, gpointer data);
void    gdownman_download_new       (GtkWidget *caller, gpointer data);
void    gdownman_project_properties (GtkWidget *caller, gpointer data);
void    gdownman_download_start     (GtkWidget *caller, gpointer data);
void    gdownman_download_stop      (GtkWidget *caller, gpointer data);
void    gdownman_download_queue     (GtkWidget *caller, gpointer data);
void    gdownman_download_delete    (GtkWidget *caller, gpointer data);
void    gdownman_download_auto      (GtkWidget *caller, gpointer data);

Project *   gdownman_get_current_project    (Gdownman *gdownman);

gboolean    gdownman_get_data               (GIOChannel *source,
                                             GIOCondition condition,
                                             gpointer data);

/* Menu and toolbar are stored here */
#include "gdownman_ui.h"

static gpointer parent_class = NULL;

GType gdownman_get_type ()
{
    static GType gdownman_type = 0;

    if (!gdownman_type)
    {
        static const GTypeInfo gdownman_info =
        {
            sizeof (GdownmanClass),
            (GBaseInitFunc) NULL,
            (GBaseFinalizeFunc) NULL,
            (GClassInitFunc) gdownman_class_init,
            NULL,           /* class_finalize */
            NULL,           /* class_data */
            sizeof (Gdownman),
            0,              /* n_preallocs */
            (GInstanceInitFunc) gdownman_init
        };

        gdownman_type = g_type_register_static (G_TYPE_OBJECT,
                                               "Gdownman",
                                               &gdownman_info,
                                               0);
    }

    return gdownman_type;
}

static void gdownman_class_init (GdownmanClass *klass)
{
    parent_class = g_type_class_peek_parent (klass);
}

static void gdownman_init (Gdownman *gdownman)
{
    gdownman->app = 0;
    gdownman->gnome_program = 0;
    gdownman->gnome_client = 0;
    gdownman->page_count = 0;
    gdownman->connected = FALSE;
    gdownman->fd = 0;
    gdownman->input_tag = 0;

    gdownman->projects = g_hash_table_new (NULL, NULL);

    gdownman->project_properties = NULL;
}

Gdownman *gdownman_new (int argc, char *argv[])
{
    Gdownman *gdownman = GDOWNMAN (g_object_new (gdownman_get_type (), NULL));

    gnome_vfs_init();

    gdownman->gnome_program = gnome_program_init("gdownman", "0.0",
                                                 LIBGNOMEUI_MODULE,
                                                 argc, argv,
                                                 GNOME_PARAM_POPT_TABLE,
                                                 NULL, NULL);

    /* initialize glade for gnome */
    glade_gnome_init ();

    gdownman->app = gnome_app_new("gdownman", "Gnome Download Manager");

	return gdownman;
}

void gdownman_delete(Gdownman *gdownman)
{
    if (gdownman->connected)
        close(gdownman->fd);
}

void gdownman_ui_build(Gdownman *gdownman)
{
	GnomeApp *app = GNOME_APP(gdownman->app);
	GtkWidget *statusbar, *hbox;

	g_signal_connect(G_OBJECT(app), "delete_event",
					G_CALLBACK(delete_event_cb), gdownman);

	gtk_window_set_default_size(GTK_WINDOW(app), 400, 350);

	hbox = gtk_hbox_new(FALSE, FALSE);
	statusbar = gtk_statusbar_new();
	gtk_statusbar_set_has_resize_grip(GTK_STATUSBAR(statusbar), TRUE);
	gtk_box_pack_start(GTK_BOX(hbox), statusbar, TRUE, TRUE, 0);
	gnome_app_set_statusbar_custom(app, hbox, statusbar);

	gnome_app_create_menus_with_data(app, menu, gdownman);
	gnome_app_install_menu_hints(app, menu);

	gnome_app_create_toolbar_with_data(app, toolbar, gdownman);

	gdownman->toolbar = GTK_WIDGET(gnome_app_get_dock_item_by_name(
								GNOME_APP(app),	GNOME_APP_TOOLBAR_NAME));
	
	gdownman->notebook = gtk_notebook_new();
	gnome_app_set_contents(app, gdownman->notebook);

    g_signal_connect(G_OBJECT(gdownman->notebook), "switch_page",
                     G_CALLBACK(gdownman_project_change), gdownman);

	gnome_app_enable_layout_config(app, TRUE);
}

void gdownman_sm_init(Gdownman *gdownman)
{
	gdownman->gnome_client = gnome_master_client();
}

void gdownman_ui_show(Gdownman *gdownman)
{
	gtk_widget_show_all(gdownman->app);
}

void gdownman_quit(GtkWidget * caller, gpointer data)
{
	Gdownman *gdownman = (Gdownman *)data;

//FIXME: save settings
	gdownman_delete(gdownman);
	gtk_main_quit();
}

static gint
delete_event_cb(GtkWidget* window, GdkEventAny* e, gpointer data)
{
	gdownman_quit(window, data);

	// Do not destroy it, we did at gdownman_quit()
	return TRUE;
}

void gdownman_project_change(GtkWidget *caller, GtkNotebookPage *page,
                             guint page_num, gpointer data)
{
    Gdownman *gdownman = (Gdownman *)data;
    Project *project;
	GtkWidget *child;

	child = gtk_notebook_get_nth_page(GTK_NOTEBOOK(gdownman->notebook),
                                      page_num);

    if (child == NULL)
        project = NULL;
    else
        project = (Project *) g_object_get_data(G_OBJECT (child), "project");

    if (project == NULL)
        return;

    if (gdownman->project_properties != NULL)
    {
        project_properties_project_set (
                PROJECT_PROPERTIES (gdownman->project_properties),
                project);
    }

    g_print("The selected project is %s\n", project_name(project));
}

void gdownman_project_add(Gdownman *gdownman, Project *project)
{
	GtkWidget *tab_label = gtk_label_new(project_name(project));
	GtkWidget *child = project_view_new(project);

    g_hash_table_insert (gdownman->projects,
                         GINT_TO_POINTER (project_id (project)),
                         project);

	g_object_set_data(G_OBJECT(child), "project", project);

	gtk_notebook_append_page(GTK_NOTEBOOK(gdownman->notebook),
							child, tab_label);

	gdownman->page_count++;

	gtk_notebook_set_show_tabs(GTK_NOTEBOOK(gdownman->notebook),
						(gdownman->page_count < 1) ? FALSE : TRUE);

	gtk_widget_show(child);
}

void gdownman_project_remove(Gdownman *gdownman, Project *project)
{
	GtkWidget *child;
    gboolean removed;
	int i;

    removed = g_hash_table_remove (gdownman->projects,
                                   GINT_TO_POINTER (project_id (project)));

    if (!removed)
        g_warning ("The project was not found in the hash table while trying to remove it.");

	for (i = 0; i < gdownman->page_count; i++)
	{
		child = gtk_notebook_get_nth_page(GTK_NOTEBOOK(gdownman->notebook), i);
		if (project == g_object_get_data(G_OBJECT(child), "project"))
		{
			gtk_notebook_remove_page(GTK_NOTEBOOK(gdownman->notebook), i);
			break;
		}
	}

	project_delete(project);
}

void gdownman_connect_local(Gdownman *gdownman)
{
	gdownman->fd = dm_client_local_connect("gdownman");

//FIXME: Try autostart before reporting error
	if (gdownman->fd < 0)
	{
		GtkWidget *dialog = gtk_message_dialog_new (GTK_WINDOW(gdownman->app),
								GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR,
								GTK_BUTTONS_OK, "%s", "There was an error while connecting to the daemon. Please, check that downmand is running.");

		gtk_dialog_run(GTK_DIALOG(dialog));
		gtk_widget_destroy(dialog);

//FIXME: start a gtk_timeout to try to connect to the local daemon (every 1s?).
		gtk_widget_set_sensitive(GTK_WIDGET(gdownman->toolbar), FALSE);
		gtk_widget_set_sensitive(GTK_WIDGET(file_menu[0].widget), FALSE);
		return;
	}

	gdownman->connected = TRUE;

    gdownman->local_channel = g_io_channel_unix_new (gdownman->fd);
    gdownman->input_tag = g_io_add_watch_full (gdownman->local_channel,
                                               G_PRIORITY_LOW,
                                               G_IO_IN,
                                               gdownman_get_data,
                                               gdownman,
                                               NULL);
	dm_command_events(gdownman->fd, 1, NULL);
	dm_command_batchprogress(gdownman->fd, 1, NULL);
	dm_command_batchstatus(gdownman->fd, 1, NULL);
}

void gdownman_project_new(GtkWidget *caller, gpointer data)
{
	Gdownman *gdownman = (Gdownman *)data;
    GString *item = g_string_new (NULL);
	GtkWidget *dialog, *label, *entry;
	const gchar *name;

    NewProject *newproject = new_project_new (gdownman->fd);

    if (newproject)
        new_project_start (newproject);
}

void gdownman_project_delete(GtkWidget *caller, gpointer data)
{
	Gdownman *gdownman = (Gdownman *)data;
	Project *project = gdownman_get_current_project(gdownman);

	dm_command_project_delete(gdownman->fd, project_id(project));
}

void gdownman_download_new(GtkWidget *caller, gpointer data)
{
	Gdownman *gdownman = (Gdownman *)data;
    Project *project = gdownman_get_current_project (gdownman);
    NewUrl *newurl;

    newurl = new_url_new (project_id (project), project->settings,
                          gdownman->fd);

    if (newurl)
        new_url_start (newurl);
}

void gdownman_project_properties(GtkWidget *caller, gpointer data)
{
	Gdownman *gdownman = (Gdownman *)data;

    if (gdownman->project_properties == NULL)
    {
        gdownman->project_properties = project_properties_new(
                                            gdownman_get_current_project(gdownman),
                                            "Project properties",
                                            GTK_WINDOW(gdownman->app));
    }

    gtk_widget_show(gdownman->project_properties);
}


void gdownman_send(Gdownman *gdownman, char *data)
{
	write(gdownman->fd, data, strlen(data) + 1);
}

Project * gdownman_get_current_project(Gdownman *gdownman)
{
    Project *project;
	int i = gtk_notebook_get_current_page(GTK_NOTEBOOK(gdownman->notebook));
	GtkWidget *child;
	
	child = gtk_notebook_get_nth_page(GTK_NOTEBOOK(gdownman->notebook), i);

    if (child == NULL)
        project = NULL;
    else
        project = (Project *)g_object_get_data(G_OBJECT(child), "project");

    return project;
}

Project *
gdownman_project_from_id (Gdownman *gdownman, guint id)
{
    Project *project;

    project = g_hash_table_lookup (gdownman->projects,
                                   GINT_TO_POINTER (id));

    return project;
}

void gdownman_select_all(GtkWidget *caller, gpointer data)
{
    Gdownman *gdownman = (Gdownman *) data;

    project_select_all (gdownman_get_current_project(gdownman));
}

void gdownman_select_done (GtkWidget *caller, gpointer data)
{
    Gdownman *gdownman = (Gdownman *) data;

    project_select_status (gdownman_get_current_project(gdownman),
                           STATUS_DONE);
}

void gdownman_select_stop (GtkWidget *caller, gpointer data)
{
    Gdownman *gdownman = (Gdownman *) data;

    project_select_status (gdownman_get_current_project(gdownman),
                           STATUS_STOPPED);
}

void gdownman_select_queue (GtkWidget *caller, gpointer data)
{
    Gdownman *gdownman = (Gdownman *) data;

    project_select_status (gdownman_get_current_project(gdownman),
                           STATUS_QUEUED);
}

void
gdownman_select_active (GtkWidget *caller, gpointer data)
{
    Gdownman *gdownman = (Gdownman *) data;

    project_select_active (gdownman_get_current_project(gdownman));
}

void
gdownman_select_matched (GtkWidget *caller, gpointer data)
{
	Gdownman *gdownman = GDOWNMAN (data);
	GtkWidget *dialog, *label, *entry, *hint_pixmap, *hint_label, *hbox;
	const gchar *text;

	dialog = gtk_dialog_new_with_buttons ("Select downloads",
					GTK_WINDOW(gdownman->app), 0,
					GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
					GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);

	label = gtk_label_new ("<big><b>File is:</b></big>");
	gtk_label_set_use_markup (GTK_LABEL(label), TRUE);
	entry = gtk_entry_new ();

    hint_pixmap = gtk_image_new_from_stock (GTK_STOCK_DIALOG_INFO,
                                            GTK_ICON_SIZE_SMALL_TOOLBAR);
    hint_label = gtk_label_new ("*.bz2 will match all bz2 files");
    hbox = gtk_hbox_new (FALSE, 5);
    gtk_container_set_border_width (GTK_CONTAINER (hbox), 10);
    gtk_container_set_border_width (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), 10);

    gtk_box_pack_start (GTK_BOX (hbox), hint_pixmap, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (hbox), hint_label, FALSE, FALSE, 0);

	gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);

	gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), label);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), entry);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), hbox);
	gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
	gtk_widget_grab_focus(entry);

	gtk_widget_show_all (dialog);

	switch(gtk_dialog_run(GTK_DIALOG(dialog)))
	{
		case GTK_RESPONSE_ACCEPT:
			text = gtk_entry_get_text(GTK_ENTRY(entry));
			if (strlen (text) > 0)
                project_select_matched (gdownman_get_current_project(gdownman),
                                        text);
			break;

		default:
	}

	gtk_widget_destroy(dialog);
}


void gdownman_download_start(GtkWidget *caller, gpointer data)
{
	Gdownman *gdownman = (Gdownman *) data;

	project_download(gdownman_get_current_project(gdownman),
					DOWNLOAD_START, gdownman->fd);
}

void gdownman_download_stop(GtkWidget *caller, gpointer data)
{
	Gdownman *gdownman = (Gdownman *) data;

	project_download(gdownman_get_current_project(gdownman),
					DOWNLOAD_STOP, gdownman->fd);
}

void gdownman_download_queue(GtkWidget *caller, gpointer data)
{
	Gdownman *gdownman = (Gdownman *) data;

	project_download(gdownman_get_current_project(gdownman),
					DOWNLOAD_QUEUE, gdownman->fd);
}

void gdownman_download_delete(GtkWidget *caller, gpointer data)
{
	Gdownman *gdownman = (Gdownman *) data;

	project_download(gdownman_get_current_project(gdownman),
					DOWNLOAD_DELETE, gdownman->fd);
}

gboolean
gdownman_get_data (GIOChannel *source, GIOCondition condition, gpointer data)
{
	Gdownman *gdownman = (Gdownman *) data;
	Project *project;
	unsigned int project_id;
	GtkWidget *dialog;
	char s[2048];
	long int i,j;
	double k,l;
	double percent;
	int id, status;
    guint maxdownloads;
    GIOStatus iostatus;
    GString *buffer = g_string_new (NULL);
    gint terminator_pos;
    GError *error = NULL;
    ProjectItem *item;
    
//FIXME: utillib func !!!!!!!!!!!!!!!!!!!

    iostatus = g_io_channel_read_line_string (source, buffer, &terminator_pos,
                                              &error);
    if ((iostatus == G_IO_STATUS_AGAIN) || (iostatus = G_IO_STATUS_ERROR))
    {
        g_string_free (buffer, TRUE);
        return TRUE;
    }

//    g_print ("GOT: %s\n", buffer->str);

    if (sscanf (buffer->str, "%li", &i) == 0)
	    return TRUE;

    switch(i)
    {
        case DM_COMMAND_NEW_ITEM:
            item = g_malloc (sizeof (ProjectItem));

            if (sscanf (buffer->str + 4, "%u %i %i %s %lf %lf",
                        &project_id,
                        &item->id,
                        &item->status,
                        s,
                        &item->downloaded,
                        &item->total) < 6)
            {
                g_free (item);
                return TRUE;
            }
            item->url = g_strdup (s);

            if (item->total > 0)
                item->percent = (item->downloaded/item->total)*100;
            else
                item->percent = 0;

            item->type = PROJECT_ITEM_NEW;

            project = gdownman_project_from_id (gdownman, project_id);

            if (project)
                project_queue_item (project, item);
            break;

        case DM_COMMAND_BATCH_PROGRESS:
            item = g_malloc (sizeof (ProjectItem));

            if (sscanf (buffer->str + 4, "%u %i %i %i %lf %lf",
                        &project_id,
                        &item->id,
                        &item->status,
                        &item->speed,
                        &item->downloaded,
                        &item->total) < 6)
            {
                g_free (item);
                return TRUE;
            }

            if (item->total > 0)
                item->percent = (item->downloaded/item->total)*100;
            else
                item->percent = 0;

            item->type = PROJECT_ITEM_UPDATE;

            project = gdownman_project_from_id (gdownman, project_id);

            if (project)
                project_queue_item (project, item);
            break;

        case DM_COMMAND_DONE_ITEM:
            item = g_malloc (sizeof (ProjectItem));

            if (sscanf (buffer->str + 4, "%u %i %i %lf %lf",
                        &project_id,
                        &item->id,
                        &item->status,
                        &item->downloaded,
                        &item->total) < 5)
            {
                g_free (item);
                return TRUE;
            }

            if (item->total > 0)
                item->percent = (item->downloaded/item->total)*100;
            else
                item->percent = 0;

            item->type = PROJECT_ITEM_DONE;
            item->speed = -1;

            project = gdownman_project_from_id (gdownman, project_id);

            if (project)
                project_queue_item (project, item);
			break;

		case DM_COMMAND_STOP_ITEM:
            item = g_malloc (sizeof (ProjectItem));

            if (sscanf (buffer->str + 4, "%u %i %i %lf %lf",
                        &project_id,
                        &item->id,
                        &item->status,
                        &item->downloaded,
                        &item->total) < 5)
            {
                g_free (item);
                return TRUE;
            }

            if (item->total > 0)
                item->percent = (item->downloaded/item->total)*100;
            else
                item->percent = 0;

            item->type = PROJECT_ITEM_STOP;
            item->speed = -1;

            project = gdownman_project_from_id (gdownman, project_id);

            if (project)
                project_queue_item (project, item);
			break;

		case DM_COMMAND_DELETE_ITEM:
            item = g_malloc (sizeof (ProjectItem));

            if (sscanf (buffer->str + 4, "%u %i",
                        &project_id,
                        &item->id) < 2)
            {
                g_free (item);
                return TRUE;
            }

            item->type = PROJECT_ITEM_DELETE;

            project = gdownman_project_from_id (gdownman, project_id);

            if (project)
                project_queue_item (project, item);
			break;

		case DM_COMMAND_BATCH_STATUS:
//FIXME: utillib func for parsing
		    if (sscanf (buffer->str + 4, "%u %li %li %lf %lf",
						&project_id, &i, &j, &k, &l) < 5)
				return TRUE;

			if (l > 0)
				percent = (k/l)*100;
			else
				percent = 0;

			project = gdownman_project_from_id(gdownman, project_id);

			if (project)
			{
                g_string_printf (buffer,
                                 "Downloading %li of %li files - %.1f%% done",
                                 i,
                                 i + j,
                                 percent);
                project_status_set (project, buffer->str);
			}
			break;

		case DM_COMMAND_PROJECT_NEW:
		    if (sscanf (buffer->str + 4, "%i %s", &id, s) < 2)
				return TRUE;

			gdownman_project_add(gdownman, project_new(id, s, gdownman->fd));

			break;

		case DM_COMMAND_PROJECT_DELETE:
		    if (sscanf (buffer->str + 4, "%i", &id) < 1)
				return TRUE;

			project = gdownman_project_from_id(gdownman, id);

			if (project)
				gdownman_project_remove(gdownman, project);
			else
				g_print("Delete project error: %i\n", id);
			break;

        case DM_COMMAND_PROJECT_PATH:
            if (sscanf (buffer->str + 4, "%i %s", &project_id, s) < 2)
                return TRUE;

            project = gdownman_project_from_id (gdownman, project_id);

            if (project)
            {
                project_auto_set (project, status);
                project_path_set (project, s);
            }
            break;

		case DM_COMMAND_PROJECT_AUTO:
		    if (sscanf (buffer->str + 4, "%i %u", &project_id, &status) < 2)
				return TRUE;

			project = gdownman_project_from_id (gdownman, project_id);

			if (project)
            {
                project_auto_set (project, status);
            }
			break;

        case DM_COMMAND_PROJECT_MAXDOWNLOADS:
		    if (sscanf (buffer->str + 4, "%i %u", &project_id, &maxdownloads) < 2)
				return TRUE;

			project = gdownman_project_from_id (gdownman, project_id);

			if (project)
            {
                project_maxdownloads_set (project, maxdownloads);
            }
			break;

		case DM_COMMAND_ERROR:
            dialog = gtk_message_dialog_new (GTK_WINDOW(gdownman->app),
                            GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_MODAL,
                            GTK_MESSAGE_ERROR,
                            GTK_BUTTONS_CLOSE,
                            "%s", buffer->str + 4);

			g_signal_connect_swapped (GTK_OBJECT (dialog), "response",
								G_CALLBACK (gtk_widget_destroy),
								GTK_OBJECT (dialog));

			gtk_widget_show(dialog);
			break;

	    default:
	        g_print ("Not handled response!!!\n");
	        break;
	}
    return TRUE;
}
