#include <unistd.h>
#include <string.h>

#include <glib/giochannel.h>
#include <glib/gstring.h>

#include <libxml/uri.h>

#include "http.h"

#define DEFAULT_HTTP_PORT 80

int http_get(int sock_fd, xmlURIPtr url, gulong file_offset,
    Linelog *linelog)
{
	GString *buffer = g_string_new(NULL);
    GString *logger;
    GError *error = NULL;

	// construct the command
	g_string_printf(buffer, "GET %s HTTP/1.0\r\n", (url->path) ? url->path : "/");

	if (file_offset)
		g_string_append_printf(buffer, "Range: bytes=%li-\r\n", file_offset);

	g_string_append_printf(buffer, "User-Agent: downmand/0.0\r\nHost: %s:%i\r\nAccept: */*\r\n\r\n", url->server, (url->port) ? url->port : DEFAULT_HTTP_PORT);

	if (write(sock_fd, buffer->str, buffer->len) == -1)
	{
		//FIXME: log the error
		perror("sending http request");
		return 1;
    }

    logger = g_string_new (NULL);
    g_string_printf (logger, "S %s", buffer->str);
    linelog_add (linelog, logger, &error);

	g_string_free(buffer, TRUE);

	return 0;
}

gint
http_get_head (GIOChannel *data_channel, gulong *size,
               Linelog *linelog)
{
	char *buf;
	GString *buffer = g_string_new(NULL);
    GString *logger;
    GError *error = NULL;
    g_io_channel_set_line_term (data_channel, NULL, -1);
	g_io_channel_set_encoding (data_channel, NULL, NULL);

	/* Read single or multiline responses */
	do {
		g_io_channel_read_line_string (data_channel, buffer, NULL, NULL);

        logger = g_string_new (NULL);
        g_string_printf (logger, "R %s", buffer->str);
        linelog_add (linelog, logger, &error);

		buf = strstr(buffer->str, "Content-Length:");
		if (buf)
		{
			buf += 16;
			if (sscanf(buf, "%li", size) > 0)
			{
                logger = g_string_new (NULL);
                g_string_printf (logger, "* Content length is %li bytes\n",
                                 *size);
                linelog_add (linelog, logger, &error);
			}
            continue;
		}

        buf = strstr(buffer->str, "Content-Range:");
        if (buf)
        {
            buf += 15;
            buf = strstr (buf, "/");
            if (buf)
            {
                buf++;
                if (sscanf(buf, "%li", size) > 0)
                {
                    logger = g_string_new (NULL);
                    g_string_printf (logger, "* File size is %li bytes\n",
                                     *size);
                    linelog_add (linelog, logger, &error);
                }
            }
            continue;
        }

    } while (strlen(buffer->str) > 2);

	g_string_free(buffer, TRUE);
	return 0;
}
